/*
** Module   : Menu.java
** Abstract : Menu abstract client implementation
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------------Description-----------------------------
** 001 VIG 20150116 Created initial version.
** 002 VIG 20150226 Added method getFirstEnabledWidget.
*/
package com.goldencode.p2j.ui.client;

import java.util.List;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.KeyListener;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.Iterables;

/**
 * <code>Menu</code> client abstract implementation class. 
 */
public abstract class Menu<O extends OutputManager<?>>
extends AbstractContainer<O>
{
   /** Menu attribute container. */
   private MenuConfig config = null; 

   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public Menu(WidgetId id, MenuConfig cfg)
   {
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, WidgetConfig cfg)
   {
      config = ConfigManager.getInstance().getActiveConfig(id);
      if (config == null)
      {
         config = new MenuConfig(id);
      }
      ConfigManager.getInstance().replaceConfig(config, cfg);
   }   

   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public MenuConfig config()
   {
      return config;
   }
   
   /**
    * Set displayed flag of this menu.
    * 
    * @param   displayed
    *          <code>true</code> means that menu is dropped down, 
    *          <code>false</code> otherwise.
    */
   public void setDisplayed(boolean displayed)
   {
      if (config.popupOnly)
      {
         config.displayed = displayed;
      }
   }

   /**
    * Check if widget is displayed.
    * 
    * @return  <code>true</code> if widget is displayed.
    */
   @Override
   public boolean isDisplayed()
   {
      if (config.popupOnly)
      {
         return config.displayed;
      }

      Widget<?> owner = screen().getRegistry().getComponent(config.ownerId);
      
      return owner == null ? false : owner.isDisplayed();
   }

   /**
    * Looks up the very first enabled widget.
    * 
    * @return  the first enabled widget or <code>null</code>
    */
   public Widget<O> getFirstEnabledWidget()
   {
      // find the first enabled widget of this frame
      Widget<O>[] widgets = widgets();
   
      for (Widget<O> comp : widgets)
      {
         if (UiUtils.hasConfig(comp) && comp.focusTraversable())
         {
            return comp;
         }
      }
   
      return null; 
   }
   
   /**
    * Set focus to specified widget.
    *
    * @param   focus
    *          Widget to set focus on.
    */
   @Override
   @SuppressWarnings("unchecked")
   public void setFocus(Widget<O> focus)
   {
      if (config.popupOnly && config.visible)
      {
         setFocusInt(focus);
         Widget<?> owner = screen().getRegistry().getComponent(config.ownerId);
         if (owner instanceof Container)
         {
            ((Container<O>) owner).setFocus(this);
         }
      }
      else
      {
         super.setFocus(focus);
      }
   }
   
   /**
    * Check if widget has focus.
    * 
    * @return  <code>true</code> if widget has focus.
    */
   @Override
   public boolean hasFocus()
   {
      if (config.popupOnly && config.visible)
      {
         Widget<?> owner = screen().getRegistry().getComponent(config.ownerId);
         return (owner == null || 
                (owner instanceof Container && ((Container<?>)owner).focus() != this)) ? 
                false : owner.hasFocus();
         
      }
      
      return super.hasFocus();
   }

   /**
    * Process key event.
    *
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      if (event.isConsumed())
      {
         return;
      }
      
      //int key = event.keyCode();
      int actionCode = event.actionCode();
      
      // menu-item choosing events processing
      if (actionCode == Keyboard.SE_VALUE_CHANGED ||
          actionCode == Keyboard.SE_CHOOSE)
      {
         hide();
         return;
      }
      
      // TODO add MENU-DROP event
      // EventManager.postEvent(new KeyInput(this, EventType.MENU_DROP_EVENT, Key.VK_ENTER));

      // mnemonic 
      if (event.keyChar() != KeyInput.CHAR_UNDEFINED)
      {
         char ch = Character.toLowerCase(event.keyChar());
         for(Widget<?> w : getWidgetsList())
         {
            if (w instanceof MenuItem)
            {
               MenuItem<?> m = (MenuItem<?>) w;
               if (m.config().mnemonic != null && // mnemonic 
                   Character.toLowerCase(m.config().mnemonic.charValue()) == ch) 
               {
                  EventManager.postEvent(new ItemEvent(m));
                  return;
               }
            }
         }
      }
      // accelerator
      else if (event.isRealKey())
      {
         int code = event.keyCode();
         for(Widget<?> w : getWidgetsList())
         {
            if (w instanceof MenuItem)
            {
               MenuItem<?> m = (MenuItem<?>) w;
               if (m.config().accelerator != null && 
                   Keyboard.keyCode(m.config().accelerator) == code)
  
               {
                  EventManager.postEvent(new ItemEvent(m));
                  return;
               }
            }
         }
      }
      else
      {
         super.processKeyEvent(event);
      }
   }
   
   /**
    * Check if this menu-item is right inside MENUBAR (first level)
    * 
    * @param    item
    *           {@link MenuItem} or {@link SubMenu}
    *           
    * @return  <code>true</code> if yes.
    */
   public static boolean isMenubarElement(Widget<?> item)
   {
      if (!(item instanceof MenuItem) && !(item instanceof SubMenu))
      {
         throw new IllegalArgumentException("Specified widget must be a menu-item or sub-menu");
      }
      
      Container<?> parent = item.parent();
      return parent instanceof Menu && !((Menu<?>)parent).config().popupOnly;
   }
   
   /**
    * Get menu widget from menu-item or sub-menu.
    * 
    * @param    item
    *           {@link MenuItem} or {@link SubMenu}
    *           
    * @return   The menu, containing specified item.
    */
   static Menu<?> getMenu(Widget<?> item)
   {
      if (!(item instanceof MenuItem) && !(item instanceof SubMenu))
      {
         throw new IllegalArgumentException("Specified widget must be a menu-item or sub-menu");
      }
      
      Widget<?> parent = item.parent();
      while (parent != null && !(parent instanceof Menu<?>))
      {
         parent = parent.parent();
      }
      
      if (parent == null)
      {
         throw new IllegalStateException("Specified widget is not contained in a menu");
      }
      
      return (Menu<?>) parent;
   }

   /**
    * Search for the next or previous focusable widget in the container and
    * set focus on it.
    *
    * @param   direction
    *          Widget search direction.
    */
   protected void focusWorker(boolean direction)
   {
      List<Widget<O>> widgets = getWidgetsList();
      Iterable<Widget<O>> iter = (direction) ?
                                     Iterables.directFrom(widgets, focus()) :
                                     Iterables.reverseFrom(widgets, focus());
                                     
       for (Widget<O> widget : iter)
       {
          if (checkWidget(widget, direction))
             return;
       }
    
       // cyclic transition to the first or last element
       Widget<O> nextWidget = direction ? widgets.get(0) : widgets.get(widgets.size() - 1);
       checkWidget(nextWidget, direction);
   }
   
   /**
    * Key listener, intended to show pop-up menu. Used for different components. 
    * Keys:
    *   ESC-U in CHUI and SHIFT-F10 in GUI
    */
   public static class PopupKeyListener 
   implements KeyListener
   {
      /**
       * This method is invoked when key is pressed.
       * 
       * @param   event
       *          Event which triggered this call.
       */
      @Override
      @SuppressWarnings({ "unchecked", "rawtypes" })
      public void onKeyPressed(KeyInput evt)
      {
         // Should we process TriggerResult here?
         
         Widget<?> src = evt.source();
         int action = (evt == null ? KeyInput.CHAR_UNDEFINED  : evt.actionCode());

         // an active pop-up menu key. TODO check for GUI
         if (action == Keyboard.KA_DEFAULT_POP_UP && src != null)
         {
            int menuId = src.config().popupMenuId;
            if (menuId != -1)
            {
               Menu<OutputManager> menu = (Menu<OutputManager>) src.screen().getRegistry().getComponent(menuId);
               if (menu != null)
               {
                  if (menu.parent() == null)
                  {
                     Window.<OutputManager>getDefaultWindow().getContentPane().add(menu);
                  }
                  
                  if (!menu.isVisible())
                  {
                     menu.setVisible(true);
                     menu.firstFocus();
                  }
                  
                  menu.repaint();
                  
                  KeyInput e = new KeyInput(menu, 
                                            EventType.KEY_PRESSED, 
                                            Keyboard.SE_MENU_DROP, 
                                            false, 
                                            false);
                  EventManager.postEvent(e);
                  
                  evt.consume();
               }
            }
         }
      }

      /**
       * This method is invoked when key is typed.
       * 
       * @param   event
       *          Event which triggered this call.
       */
      @Override
      public void onKeyTyped(KeyInput event)
      {
         // Nothing
      }
   }
}
