/*
** Module   : WindowChuiImpl.java
** Abstract :
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140521 First version. Extracted from WindowImpl class.
** 002 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Removed the WindowImpl class and moved code here - all base behavior is in the
**                  Window class.
**                  Refactoring to allow multi-window support.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 MAG 20141009 Added INSERT indicator widget which is ChUI specific only.
** 006 HC  20141021 Implemented window move/position support.
** 007 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 008 CA  20150219 Hiding the ChUI window removes the FOCUS reference (sets it to unknown).  The
**                  ChUI WINDOW's VISIBLE attribute can not be set to FALSE.
** 009 VIG 20150121 Added MENU drawing.
** 010 VIG 20150226 Added menu parent setting before drawing method call.
*/

package com.goldencode.p2j.ui.client.chui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.chui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

@SuppressWarnings("unchecked")
public class WindowChuiImpl
extends Window<ChuiOutputManager>
{
   /** Renders the status of the current insertion mode. */
   StatusLine<ChuiOutputManager> insertModeLabel = null;
   
   /**
    * Hide base class public constructor.
    * 
    * @param    id
    *           The widget's ID.
    */
   public WindowChuiImpl(WidgetId id)
   {
      super(id);

      init();
   }
   
   /**
    * Construct a window widget based on a specific configuration.
    * 
    * @param   config
    *          Window configuration.
    * @param   id
    *          The widget's ID.
    */
   public WindowChuiImpl(WidgetId id, WidgetConfig config)
   {
      super(id, (WindowConfig) config);
   }

   /**                                     
    * Get dimension of screen area available for the placement of frames.
    * 
    * @return  Instance of <code>Dimension</code> which describes size of the
    *          available screen area.
    */
   public Dimension getScreenDimension()
   {
      Dimension sz = dimension();
      return new Dimension(sz.width, sz.height - RESERVED_LINES);
   }
   
   /**
    * Get messages area rectangle by message line index.
    *
    * @param   line
    *          message line specification: either 0-based line number for
    *          specific line or -1 for the whole message area.
    * @return  Instance of <code>Rectangle</code> which describes the message
    *          area in screen coordinates.
    */
   public Rectangle getMessageLineRectangle(int line)
   {
      Dimension sz = dimension();
      double    y0 = sz.height - RESERVED_LINES;
      double    y1 = sz.height;
      double    x1 = sz.width - 1;
      
      if (line < 0)
      {
         y1 -= 2;
      }
      else
      {
         y0 += line;
         y1 = y1 - RESERVED_LINES + line; 
      }
         
      return new Rectangle(y0, 0, y1, x1);
   }

   /**
    * Draw all the components in this window, and request the keyboard focus.
    * <p>
    * If in suspend mode, only draw the status line.  This mode can be set
    * via {@link #suspend}.
    */
   @Override
   public void draw()
   {
      if (config.menubarId != 0)
      {
         Widget<ChuiOutputManager> menu = screen().getRegistry().getComponent(config.menubarId);
         
         if (menu != null)
         {
            if (menu.parent() == null)
            {
               getContentPane().add(menu);
            }
         
            menu.setVisible(true);
         }
      }

      super.draw();
   }

   /**
    * Get status area rectangle.
    *
    * @return  Instance of <code>Rectangle</code> which describes the status
    *          area in screen coordinates.
    */
   public Rectangle getStatusLineRectangle()
   {
      Dimension sz = dimension();
      double    y  = sz.height - 1;
      double    x  = sz.width - 1;
      
      return new Rectangle(y, 0, y, x);
   }

   /**
    * When the window is no longer visible, special processing is required.  This includes hiding 
    * all frames and clearing message and status.
    * 
    * @param    visible
    *           The frame's visibility state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      if (visible) 
      {
         _setVisible(visible);

         return;
      }

      // the ChUI window can't have its VISIBLE attribute set to FALSE
      _setVisible(true);
      
      String[] content = getMessageArea().getMessageTexts();

      Widget<ChuiOutputManager> focus = currentFocus();

      ThinClient tc = ThinClient.getInstance();
      
      tc.hideAll(true, UiUtils.getWidgetIdAsInt(this));

      if (focus != null && tc.getTriggerNesting() > 0)
      {
         Frame<ChuiOutputManager> frame = (Frame<ChuiOutputManager>) UiUtils.locateFrame(focus);

         if (frame != null)
            frame.setFocusReset(true);
      }

      // no more focus - is lost forever!
      setFocus(null);

      if (content[0] == null && content[1] == null)
         return;

      // message lines were not empty, so simulate hiding of message lines
      // instead of wiping out their content
      if (content[0] != null)
         content[0] = "";
      if (content[1] != null)
         content[1] = "";

      setMessageText(content, null);
   }
   
   /**
    * Reset Window instance state.
    */
   protected void reset()
   {
      setSize(1, 1);

      super.reset();      

      // INSERT indicator is ChUI specific only.
      insertModeLabel = (StatusLine<ChuiOutputManager>) new MessageLine(WidgetId.nextID());      
   }

   /**
    * Initialize window
    */
   protected void init()
   {
      add(contentPane);
      setLayout(null);
      
      reset();

      WindowConfig config = config();

      BasePrimitives direct = screen().getDirectOutput();
      
      Dimension screenSize = new Dimension(direct.screenWidth(), direct.screenHeight());
      setSize(screenSize);

      getContentPane().setSize(screenSize);
      
      // Message area with variable lines of the text
      messageArea.attach(this);
      
      // status line with default text
      statusLine.attach(this);

      // "Insert" label 
      getContentPane().add(insertModeLabel);
      insertModeLabel.setLocation(statusLine.width(), screenSize.height - 1);
      insertModeLabel.setText(null);
      insertModeLabel.setLength(INSERT_TEXT.length());
      insertModeLabel.setColor(Color.NORMAL);
      
      // Initialization of the default values for attributes
      // which are dependent from the code in this method
      // Size related attributes
      config.screenLines = Coordinate.scale(screenSize.height - RESERVED_LINES);
      config.fullHeightChar = Coordinate.scale(screenSize.height - 4);
      config.fullWidthChar = Coordinate.scale(screenSize.width);
      config.heightChars = Coordinate.scale(screenSize.height); // direct access
      config.widthChars = Coordinate.scale(screenSize.width); // direct access
      
      isRunning = true;
   }

   /**
    * Set the contents of the status line. This method is private and
    * called at service level by another static public method within the
    * class.  The status line output is truncated to a maximum of 63
    * characters, only if the <code>noLengthLimit</code> flag is not set.
    * 
    * @param   status
    *          The new string to be displayed inside the status line area.
    * @param   noLengthLimit
    *          flag indicating the status message should not be trimmed
    */
   protected void setStatusLineText(String status, boolean noLengthLimit)
   {
      int trimSize = (int) (noLengthLimit ? dimension().width : 63);
      
      // change the text
      String trim = status.length() > trimSize ? status.substring(0, trimSize)
                                               : status;
   
      getContentPane().moveToTop(statusLine);
      statusLine.setText(trim);
      statusLine.repaint();
   }

   /**
    * The private function to set up the global insert mode value.
    * 
    * @param   insMode
    *          New integer value for insert mode.
    * @param   force
    *          Whether we have to repaint the indicator even if the new mode
    *          is the same as the old one.
    */
   protected void setInsMode(int insMode, boolean force)
   {
      if (insMode == insertMode && !force)
         return;
      
      if (insMode == INSERT_MODE_INS)
      {
         if (insertMode == INSERT_MODE_OFF || insertMode == INSERT_MODE_OVR)
            insertModeLabel.setText(INSERT_TEXT);
      }
      else if (insMode == INSERT_MODE_OVR)
      {
         if (insertMode == INSERT_MODE_OFF || insertMode == INSERT_MODE_INS)
            insertModeLabel.setText(null);
      }
      else if (insMode == INSERT_MODE_OFF)
      {
         insertModeLabel.setText(null);
      }
      else
         return;
      
      getContentPane().moveToTop(insertModeLabel);
      
      // save the bitmap, so it can be restored
      ScreenBitmap bitmap = getAndResetBitmap();
      insertModeLabel.draw();
      sync();
      // restore the screen
      setBitmap(bitmap);
      
      insertMode = insMode;
      insertInAmbiguousState = false;
   }
   
   /**
    * Enables / Disables the "special components" (status line, insert mode indicator,
    * message lines, temp text components) and restores visibility to the
    * list of all previously visible top-level components.
    *
    * This is part of the switching of the client to the interactive mode.
    * 
    * @param   visibility
    *          <code>true</code> enable <code>false</code> disable.
    */
   @Override
   protected void setVisibility(boolean visibility)
   {
      super.setVisibility(visibility);
      insertModeLabel.forceVisibility(visibility);
   }   
}
